package be.ac.ulb.mlg.utils.measure;

/*
 * The MIT License (MIT)
 * 
 * Copyright (c) 2013 Jean-Sebastien Lerat (Jean-Sebastien.Lerat@ulb.ac.be)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/**
 * 
 * @author Jean-Sebastien Lerat (Jean-Sebastien.Lerat@ulb.ac.be)
 * @version 1.00, 24/06/2013
 * @see <a href="http://dx.doi.org/10.1093%2Fbiomet%2F30.1-2.81">A New Measure of Rank Correlation (Kendall, M)</a>
 */

import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.min;
import be.ac.ulb.mlg.utils.Measure;
import be.ac.ulb.mlg.utils.MeasureUtils;

/**
 * Using Tau-b (adjustments for ties)
 * Knight's Algorithm
 */
public class Kendall implements Measure{

	@Override
	public double measure(final double[] A, final double[] B,boolean[] computable){
		final double[] rankA	= new double[A.length];
		final double[] rankB	= new double[B.length];
		final int[] rankJoint	= new int[A.length];
		
		final int n1 = MeasureUtils.naturalRank(computable,A,rankA,rankJoint);
		final int n2 = MeasureUtils.naturalRank(computable,B,rankB);
		
		int count = rankA.length,tieCount=0;
		int index,n3=0;

		//compute the joint ties (n3)
		if(computable != null){
			//when no computable, stop because no computable are put to the end of positions array (during the sort)
			for(index=0;index<rankA.length && computable[rankJoint[index]];index++){
				if( (rankA[rankJoint[index]]-(int)rankA[rankJoint[index]])<1E-10 || (rankB[rankJoint[index]]-(int)rankB[rankJoint[index]])<1E-10 ){
			    	if (tieCount > 0){
			    		n3			+= tieCount*(tieCount+1);
			    		tieCount	=  0;
			    	}
				} else tieCount++;
			}
			count = index;//update the size of vectors, size-not computable
		}else{
			for(index=0;index<rankA.length;index++){
				if( (rankA[rankJoint[index]]-(int)rankA[rankJoint[index]])<1E-10 || (rankB[rankJoint[index]]-(int)rankB[rankJoint[index]])<1E-10 ){
			    	if (tieCount > 0){
			    		n3			+= tieCount*(tieCount+1);
			    		tieCount	=  0;
			    	}
				} else tieCount++;
			}
		}
		if(tieCount > 0) n3	+= tieCount*(tieCount+1);//last tie if needed
		n3 >>= 1;

		final int n0	= count*(count-1)>>1;
		if(n0 == n1 || n0 == n2) return 1.0;//MAX correlation value
		
		//sort rankB (only indicies) based on the sorting of rankA (rankJoint)
		final double numerator		= n0-n1-n2+n3 - (mergesortCounting(computable,rankJoint,rankB)<<1);//nc -nd = n0 - n1 - n2 + n3 - 2*S(y) 
		final double denominator	= exp(0.5*(log(n0-n1)+log(n0-n1)));//underflow (Numerical Analysis course) ( exp(0.5*log(x)) <=> sqrt(x) )
		
		if(denominator<1e-16)return ERROR_VALUE;
		
		return numerator/denominator;
	}
	
	private static int mergesortCounting(boolean[] computable,int[] toSort,double[] comparable){
		int[] swapper = new int[toSort.length];
		int swapCount = 0;
		int index,leftTape,rightTape,end;
		int ptrLeftTape,ptrRightTape,position;
		
		for(int tapeSize = 1; tapeSize < toSort.length; tapeSize <<= 1){
			for(index = 0; index < toSort.length; index += tapeSize<<1){
				leftTape	= index;
				rightTape	= min(index+tapeSize,toSort.length);
				end			= min(rightTape+tapeSize,toSort.length);
				
				ptrLeftTape		= leftTape;
				ptrRightTape	= rightTape;

				for(position = leftTape; position < end; position++){
					if(ptrLeftTape >= rightTape || (ptrRightTape < end && comparable[toSort[ptrLeftTape]] > comparable[toSort[ptrRightTape]])){
						swapper[position]	=  toSort[ptrRightTape++];
						swapCount			+= rightTape - ptrLeftTape;
					}else{
						swapper[position] = toSort[ptrLeftTape++];
					}
					
					/*
					// this is the detailed version condition by condition of the algorithm
					if(ptrLeftTape<rightTape){
						//left tape are not yet ended
						if(ptrRightTape<end){
							//the right tape is not ended, we must still choose between the right and the left elements
							if(toSort[ptrLeftTape] > toSort[ptrRightTape]){
								//the left element is greater than the right
								//copy the right element before the left (the right and increase the ptr of the right)
								swapper[position]	=  toSort[ptrRightTape++];
								swapCount			+= rightTape - ptrLeftTape;
							}else{
								//the left element is already smaller than the right.
								//just copy it
								swapper[position] = toSort[ptrLeftTape++];
							}
						}else{
							//right tape processing ended, just copy the left tape
							swapper[position] = toSort[ptrLeftTape++];
						}
					}else{
						//left tape processing ended, just copy the right tape (right tape is logically ended)
						swapper[position] = toSort[ptrRightTape++];
					}*/
				}
				
			}
			
			final int[] tmp	= toSort;
			toSort		= swapper;
			swapper	= tmp;
		}
		
		return swapCount;
	}
	 
	@Override
	public boolean hasNativeImplementation() {
		return true;
	}
	@Override
	public boolean requireDataTransformation() {
		return false;
	}

	@Override
	public void transform(double[][] data) {
		//nothnig to do
	}
}
